/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "gcc-compat.h"
#include "para.h"
#include "list.h"
#include "filter.h"
#include "error.h"

DEFINE_FILTER_ARRAY(filters);

int filter_init(struct filter *all_filters)
{
	struct filter *f;
	int ret;

	for (f = all_filters; f->name; f++) {
//		PARA_DEBUG_LOG("initializing %s filter\n", f->name);
		ret = f->init(f);
		if (ret < 0)
			return ret;
	}
	return 1;
}

int close_filter_callback(struct filter_callback *fcb)
{
	PARA_NOTICE_LOG("closing filter_callback %p, data: %p\n", fcb, fcb->data);
	list_del(&fcb->node);
	fcb->close(fcb);
	return 1;
}

static void close_callbacks(struct filter_node *fn)
{
	struct filter_callback *fcb, *tmp;

	list_for_each_entry_safe(fcb, tmp, &fn->callbacks, node)
		close_filter_callback(fcb);
}

static void call_callbacks(struct filter_node *fn, char *inbuf, size_t inlen,
	char *outbuf, size_t outlen)
{
	struct filter_callback *fcb, *tmp;
	list_for_each_entry_safe(fcb, tmp, &fn->callbacks, node) {
		int ret;
		if (inlen && fcb->input_cb) {
			ret = fcb->input_cb(inbuf, inlen, fcb);
			if (ret < 0) {
				close_filter_callback(fcb);
				continue;
			}
		}
		if (!outlen || !fcb->output_cb)
			continue;
		ret = fcb->output_cb(outbuf, outlen, fcb);
		if (ret < 0)
			close_filter_callback(fcb);
	}
}

int filter_io(struct filter_chain_info *fci)
{
	struct filter_node *fn;
	char *ib = fci->inbuf;
	size_t *loaded = fci->in_loaded;
	int converted = 0;

	list_for_each_entry(fn, &fci->filters, node) {
		int ret;
		if (*loaded && fn->loaded < fn->bufsize) {
			size_t old_fn_loaded = fn->loaded;
			PARA_DEBUG_LOG("fc %p loaded: %d, calling %s convert\n", fci, *loaded, fn->filter->name);
			ret = fn->filter->convert(ib, *loaded, fn, fci);
			if (ret < 0) {
				if (!fci->error)
					fci->error = -ret;
				return ret;
			}
			call_callbacks(fn, ib, ret, fn->buf + old_fn_loaded, fn->loaded - old_fn_loaded);
			*loaded -= ret;
			converted += ret;
			if (*loaded && ret) {
				PARA_DEBUG_LOG("moving %d bytes in input buffer for %s filter\n",
					*loaded,  fn->filter->name);
				memmove(ib, ib + ret, *loaded);
			}
		}
		ib = fn->buf;
		loaded = &fn->loaded;
	}
//	PARA_DEBUG_LOG("loaded: %d\n", *loaded);
	return converted;
}


void close_filters(struct filter_chain_info *fci)
{
	struct filter_node *fn, *tmp;

	if (!fci)
		return;
	PARA_DEBUG_LOG("closing filter chain %p\n", fci);
	list_for_each_entry_safe(fn, tmp, &fci->filters, node) {
		PARA_NOTICE_LOG("closing %s filter callbacks (fci %p, fn %p)\n", fn->filter->name, fci, fn);
		close_callbacks(fn);
		PARA_NOTICE_LOG("closing %s filter (fci %p, fn %p)\n", fn->filter->name, fci, fn);
		fn->filter->close(fn);
		list_del(&fn->node);
		free(fn);
	}
}

static void *parse_filter_args(int filter_num, char *options)
{
	struct filter *f = &filters[filter_num];
	int i, argc = 2;
	char *dummy_args[] = {"", "", NULL};
	char **argv;

//	PARA_DEBUG_LOG("%s, options: %s, parser: %p\n", f->name,
//		options? options : "(none)", f->parse_config);
	if (!f->parse_config)
		return options ? NULL : f;
	if (options) {
//		PARA_DEBUG_LOG("%s options: %s\n", name, options);
		argc = split_args(options, &argv, ':');
//		PARA_DEBUG_LOG("argc = %d, argv[0]: %s\n", argc, argv[0]);
		for (i = argc; i >= 0; i--)
			argv[i + 1] = argv[i];
		argc += 2;
		return f->parse_config(argc, argv);
	}
	return f->parse_config(2, dummy_args);
}

struct filter_node *check_filter_arg(char *fa, int *filter_num)
{
	int j;

//	PARA_DEBUG_LOG("arg: %s\n", fa);
	*filter_num = -E_BAD_FILTER_OPTIONS;
	for (j = 0; filters[j].name; j++) {
		const char *name = filters[j].name;
		size_t len = strlen(name);
		char c;
		if (strlen(fa) < len)
			continue;
		if (strncmp(name, fa, len))
			continue;
		c = fa[len];
		if (c && c != ':')
			continue;
		if (c && !filters[j].parse_config)
			return NULL;
		*filter_num = j;
		return parse_filter_args(j, c? fa + len + 1: NULL);
	}
	*filter_num = -E_UNSUPPORTED_FILTER;
	return NULL;
}

